"""Exception classes relating to calling the Geneweaver API."""
from geneweaver.core.exc import GeneweaverError


class GeneweaverAPIError(GeneweaverError):
    """Base class for exceptions in the Geneweaver API."""

    pass



"""CLI for parsing data files."""
from pathlib import Path

import typer
from geneweaver.client.parser import general
from geneweaver.core.parse import xlsx
from geneweaver.core.parse.exceptions import EmptyFileError, UnsupportedFileTypeError
from rich import print
from rich.console import Console
from rich.table import Table

cli = typer.Typer()
console = Console()


@cli.command()
def get_headers(file_path: Path, sheet: str = None) -> None:
    """Get the headers from a data file."""
    headers = []
    try:
        headers = general.get_headers(file_path, sheet_name=sheet)
    except (EmptyFileError, ValueError, UnsupportedFileTypeError) as e:
        print(e)
        raise typer.Exit(code=1) from e

    for header in headers:
        print(header)


@cli.command()
def preview(file_path: Path, rows_to_read: int = 5, sheet: str = None) -> None:
    """Preview the data in a data file."""
    rows = []
    try:
        headers, headers_idx = general.get_headers(file_path, sheet_name=sheet)
        rows = general.data_file_to_dict_n_rows(
            file_path, rows_to_read, headers_idx, sheet_name=sheet
        )
    except ValueError as e:
        print("File is empty.")
        raise typer.Exit(code=1) from e
    except (EmptyFileError, UnsupportedFileTypeError) as e:
        print(e)
        raise typer.Exit(code=1) from e

    table = Table(*headers)
    for row in rows:
        table.add_row(*(str(r) for r in row.values()))
    console.print(table)


@cli.command()
def get_sheet_names(file_path: Path) -> None:
    """Get the sheet names from an Excel file."""
    try:
        names = xlsx.get_sheet_names(file_path)
        print(names)
    except ValueError as e:
        print(e)
        raise typer.Exit(code=1) from e



"""The main entrypoint to the GeneWeaver CLI client."""
# ruff: noqa: B008
import pkg_resources
import typer

from geneweaver.client.cli import parser
from geneweaver.client.cli import auth

cli = typer.Typer(no_args_is_help=True)
cli.add_typer(parser.cli, name="parse")
cli.add_typer(auth.cli, name="auth")


def version_callback(version: bool) -> None:
    """Print the version of the GeneWeaver CLI client."""
    if version:
        version = pkg_resources.get_distribution("geneweaver-client").version
        typer.echo(f"GeneWeaver CLI client (gweave) version: {version}")
        raise typer.Exit(code=0)


@cli.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", callback=version_callback),
) -> None:
    """GeneWeaver CLI client."""
    pass



"""Authentication CLI commands."""
import typer

from geneweaver.client.user_config import get_auth_token
from geneweaver.client import auth
from geneweaver.client.exceptions import AuthenticationError

cli = typer.Typer()


@cli.command()
def login(reauth: bool = typer.Option(False, "--reauth")) -> None:
    """Runs the device authorization flow.

    :param reauth: Force a re-authentication
    """
    if not reauth and get_auth_token():
        print("You are already logged in")
        raise typer.Exit(code=1)

    try:
        auth.login()
    except AuthenticationError as e:
        print(e)
        raise typer.Exit(code=1)



"""A module that marshals access to specific file type parsing."""
from typing import List, Optional

from geneweaver.core.parse import csv, utils, xlsx
from geneweaver.core.parse.exceptions import UnsupportedFileTypeError
from geneweaver.core.types import DictRow, StringOrPath


def get_headers(file_path: StringOrPath, sheet_name: Optional[str] = None) -> List[str]:
    """Retrieve the header row from a CSV or Excel file.

    This function first determines the file type (CSV or Excel) and then uses
    the appropriate helper function to get the headers.

    :param file_path: The path to the file.
    :param sheet_name: Name of the sheet to read from (for Excel files).
                       If not provided, the function will read from the active sheet.
                       This argument is ignored for CSV files.

    :returns: A list of strings representing the header row of the file.

    :raises ValueError: If the file type is neither CSV nor Excel (.xlsx).
    """
    file_type = utils.get_file_type(file_path)

    if file_type == "csv":
        data, _ = csv.get_headers(file_path)

    elif file_type == "xlsx":
        data, _ = xlsx.get_headers(file_path, sheet_name)

    else:
        raise UnsupportedFileTypeError(f"Unsupported file type: {file_type}")

    return data


def data_file_to_dict(
    file_path: StringOrPath, sheet_name: Optional[str] = None
) -> List[DictRow]:
    """Parse a CSV or Excel file into a list of dictionaries.

    Parse a CSV or Excel file into a list of dictionaries, with keys as column names and
     values as column values.

    :param file_path: The file path to the CSV or Excel file.
    :param sheet_name: Name of the sheet to read from (for Excel files). If not
    provided, the function will read from the active sheet. Ignored for CSV files.

    :returns: A list of dictionaries, where each dictionary represents a row from the
    CSV or Excel file.
    """
    file_type = utils.get_file_type(file_path)

    if file_type == "csv":
        data = csv.read_to_dict(file_path)

    elif file_type == "xlsx":
        data = xlsx.read_to_dict(file_path, sheet_name)

    else:
        raise UnsupportedFileTypeError(f"Unsupported file type: {file_type}")

    return data


def data_file_to_dict_n_rows(
    file_path: StringOrPath,
    n: int,
    start_row: int = 0,
    sheet_name: Optional[str] = None,
) -> List[DictRow]:
    """Parse n lines of a CSV or Excel file into a list of dictionaries.

    Parse a CSV or Excel file into a list of dictionaries, with keys as column names and
    values as column values.

    :param file_path: The file path to the CSV or Excel file.
    :param n: The number of rows of data to return.
    :param start_row: The row to start reading from. Defaults to 0.
    :param sheet_name: Name of the sheet to read from (for Excel files). If not
    provided, the function will read from the active sheet. Ignored for CSV files.

    :returns: A list of dictionaries, where each dictionary represents a row from the
    CSV or Excel file.
    """
    file_type = utils.get_file_type(file_path)

    if file_type == "csv":
        data = csv.read_to_dict_n_rows(file_path, n, start_row)

    elif file_type == "xlsx":
        data = xlsx.read_to_dict_n_rows(file_path, n, start_row, sheet_name)

    else:
        raise UnsupportedFileTypeError(f"Unsupported file type: {file_type}")

    return data



"""GeneWeaver Client configuration module."""
from typing import Optional, Type

from pydantic import BaseModel, BaseSettings, validator


class LegacyClientPaths(BaseModel):
    """Configuration schema for legacy client paths."""

    KEYWORD_SEARCH_GUEST: str = "/api/get/search/bykeyword/{apikey}/{search_term}/"

    GET_GENESETS_BY_GENE_REF_ID: str = (
        "/api/get/geneset/bygeneid/{apikey}/{gene_ref_id}/{gdb_name}/"
    )
    GET_GENESETS_BY_GENE_REF_ID_HOMOLOGY: str = (
        "/api/get/geneset/bygeneid/{apikey}/{gene_ref_id}/{gdb_name}/homology"
    )
    GET_GENESET_BY_USER: str = "/api/get/geneset/byuser/{apikey}/"
    GET_ONTOLOGY_BY_GENESET_ID: str = (
        "/api/get/ontologies/bygeneset/{apikey}/{geneset_id}/"
    )

    GET_GENESET_BY_ID: str = "/api/get/geneset/byid/{genesetid}/"
    GET_GENES_BY_GENESET_ID: str = "/api/get/genes/bygenesetid/{geneset_id}/"

    GET_GENE_BY_GENE_ID: str = "/api/get/gene/bygeneid/{gene_id}/"
    GET_PROJECTS_BY_USER: str = "/api/get/project/byuser/{apikey}/"
    GET_GENESET_BY_PROJECT_ID: str = (
        "/api/get/geneset/byprojectid/{apikey}/{project_id}/"
    )
    GET_PROBES_BY_GENE: str = "/api/get/probes/bygeneid/{apikey}/{gene_ref_id}/"
    GET_PLATFORM_BY_ID: str = "/api/get/platform/byid/{apikey}/{platform_id}/"
    GET_SNP_BY_GENE_ID: str = "/api/get/snp/bygeneid/{apikey}/{gene_ref_id}/"
    GET_PUBLICATION_BY_ID: str = "/api/get/publication/byid/{apikey}/{publication_id}/"
    GET_SPECIES_BY_ID: str = "/api/get/species/byid/{apikey}/{species_id}/"
    GET_RESULTS_BY_USER: str = "/api/get/results/byuser/{apikey}/"
    GET_RESULT_BY_TASK_ID: str = "/api/get/result/bytaskid/{apikey}/{task_id}/"
    GET_GENE_DATABASE_BY_ID: str = (
        "/api/get/genedatabase/byid/{apikey}/{gene_database_id}/"
    )

    ADD_PROJECT_BY_USER: str = "/api/add/project/byuser/{apikey}/{project_name}/"
    ADD_GENESET_TO_PROJECT: str = (
        "/api/add/geneset/toproject/{apikey}/{project_id}/{geneset_id}/"
    )
    ADD_GENESET_BY_USER: str = "/api/add/geneset/byuser/{apikey}/"

    DELETE_GENESET_FROM_PROJECT: str = (
        "/api/delete/geneset/fromproject/{apikey}/{project_id}/{geneset_id}/"
    )

    ADD_PROJECT_BY_NAME: str = "/addProjectByName"

    CREATE_TEMP_GENESET_ORIGINAL: str = "/createtempgeneset_original"
    CREATE_TEMP_GENESET_LARGE: str = "/createtempgeneset_large"


class ClientSettings(BaseSettings):
    """Settings class for GeneWeaver Client."""

    AUTH_DOMAIN = "geneweaver.auth0.com"
    AUTH_CLIENT_ID = "tvcvnbBJQr15jFds8ZkwgHoqRJ0IuGJC"
    AUTH_ALGORITHMS = ["RS256"]
    AUTH_SCOPES = ["openid", "profile", "email"]

    API_HOST: str = "https://geneweaver.org"
    API_PATH: str = "/api/v2"

    API_URL: Optional[str] = None

    LEGACY_PATHS: LegacyClientPaths = LegacyClientPaths()

    @validator("API_URL")
    def validate_api_url(
        cls: Type["ClientSettings"], v: Optional[str], values: dict  # noqa: N805
    ) -> str:
        """Construct the API URL if not explicitly set."""
        if not v:
            return values["API_HOST"] + values["API_PATH"]
        return v

    API_KEY: Optional[str] = None

    class Config:
        """Settings configuration."""

        env_file = ".env"
        env_prefix = "GW_CLIENT_"


settings = ClientSettings()



import requests, zipfile, io
from pathlib import Path
import pandas as pd


class BaseDataset:

    URL = None
    DS_FOLDER = None
    UNZIPPED_LOC = None

    def __init__(self, base_folder: str = 'data'):
        self.base_folder = Path(base_folder)

    @property
    def dataset_path(self):
        return self.base_folder / self.DS_FOLDER / self.UNZIPPED_LOC

    def download_zip_file(self, redownload: bool = False) -> None:
        """Download the zip file from the dataset's URL."""
        if not redownload and self.dataset_path.is_file():
            return None

        response = requests.get(self.URL)
        response.raise_for_status()

        z = zipfile.ZipFile(io.BytesIO(response.content))

        z.extractall(self.base_folder / self.DS_FOLDER)

    def as_pandas(self) -> pd.DataFrame:
        """Return the dataset as a Pandas DataFrame."""
        return self._pandas_read_f(
            self.dataset_path,
            skiprows=self.dataset_skip_rows
        )


"""Functions that wrap the GeneWeaver API on /genesets endpoints."""
from typing import Optional

from geneweaver.client.api.utils import sessionmanager
from geneweaver.client.config import settings
from geneweaver.core.schema import geneset as geneset_schema

ENDPOINT = "/genesets/"


def post(
    token: str, geneset: geneset_schema.GenesetUpload
) -> Optional[geneset_schema.Geneset]:
    """Create a new Geneset, with the genes specified in the request body."""
    with sessionmanager() as session:
        resp = session.post(
            settings.API_URL + ENDPOINT,
            json=geneset.dict(),
            headers={"Authorization": token},
        )
    return geneset_schema.Geneset(**resp.json())


def post_batch(
    token: str, geneset: geneset_schema.BatchUpload
) -> geneset_schema.Geneset:
    """Create a new Geneset using a batch upload file."""
    with sessionmanager() as session:
        resp = session.post(
            settings.API_URL + ENDPOINT + "batch",
            json=geneset.dict(),
            headers={"Authorization": token},
        )
    return geneset_schema.Geneset(**resp.json())


def get(token: str, geneset_id: int) -> geneset_schema.Geneset:
    """Get a Geneset by ID."""
    with sessionmanager() as session:
        resp = session.get(
            settings.API_URL + ENDPOINT + str(geneset_id),
            headers={"Authorization": token},
        )
    return geneset_schema.Geneset(**resp.json())



"""Dataset definition for NCI60's: DNA__Combined_aCGH_gene_summary.xls"""
import pandas as pd
from geneweaver.client.datasets.base import BaseDataset


class DNACombinedaCGHGeneSummary(BaseDataset):

    URL = "https://discover.nci.nih.gov/cellminer/download/processeddataset/" \
          "nci60_DNA__Combined_aCGH_gene_summary.zip"
    DS_FOLDER = "nci60_DNA__Combined_aCGH_gene_summary"
    UNZIPPED_LOC = "output/DNA__Combined_aCGH_gene_summary.xls"

    def __init__(self, base_folder: str = 'data'):
        super().__init__(base_folder)
        self.download_zip_file()
        self.dataset_skip_rows = 10
        self._pandas_read_f = pd.read_excel

    def entrez_ids(self) -> pd.DataFrame:
        """Return the dataset's entrez ids."""
        return self.as_pandas().iloc[:, 1]

    def gene_names(self) -> pd.DataFrame:
        """Return the dataset's gene names."""
        return self.as_pandas().iloc[:, 0]

    def intensity_values(self) -> pd.DataFrame:
        """Return the dataset's intensity values."""
        return self.as_pandas().iloc[:, 6:]



"""Legacy API functions.

Functions that call the legacy GeneWeaver Flask API that's built in to the
py3-geneweaver-website repo.
"""
from typing import Optional

import requests
from geneweaver.client.config import ClientSettings
from geneweaver.core.schema import legacy_api


def add_geneset_by_user(
    settings: ClientSettings,
    geneset: legacy_api.AddGenesetByUser,
    publication: Optional[legacy_api.AddGenesetByUserPublication] = None,
) -> requests.Response:
    """Add a geneset to GeneWeaver using the legacy API."""
    url = settings.API_HOST + settings.LEGACY_PATHS.ADD_GENESET_BY_USER.format(
        apikey=settings.API_KEY
    )
    request_data = geneset.dict(exclude={"publication"})
    if publication is not None:
        request_data.update(publication.dict())
    response = requests.post(url, json=request_data)
    return response


def create_temp_geneset_original(
    # TODO: The `geneset` argument should be typed! Remove noqa when this is fixed.
    settings: ClientSettings,
    geneset,  # noqa: ANN001
) -> requests.Response:
    """Create a temporary geneset using the legacy API."""
    url = settings.API_HOST + settings.LEGACY_PATHS.CREATE_TEMP_GENESET_ORIGINAL
    response = requests.post(url, json=geneset.dict())
    return response


def add_project_by_user(
    settings: ClientSettings, project_name: str
) -> requests.Response:
    """Add a project to GeneWeaver using the legacy API."""
    url = settings.API_HOST + settings.LEGACY_PATHS.ADD_PROJECT_BY_USER.format(
        apikey=settings.API_KEY, project_name=project_name
    )
    response = requests.get(url)
    return response


def add_geneset_to_project(
    settings: ClientSettings, geneset_id: int, project_id: int
) -> requests.Response:
    """Add a geneset to a project using the legacy API."""
    url = settings.API_HOST + settings.LEGACY_PATHS.ADD_GENESET_TO_PROJECT.format(
        apikey=settings.API_KEY, geneset_id=geneset_id, project_id=project_id
    )
    response = requests.get(url)
    return response



"""Function for interacting with the Geneweaver API."""
import requests
from geneweaver.client.core.config import settings


def get_genesets(access_token: str) -> list:
    """Get all visible genesets"""
    return requests.get(
        settings.api_v3_path() + "/genesets",
        headers={"Authorization": f"Bearer {access_token}"},
    ).json()


def get_geneset(access_token: str, geneset_id: str):
    """Get a specific geneset"""
    return requests.get(
        settings.api_v3_path() + f"/genesets/{geneset_id}",
        headers={"Authorization": f"Bearer {access_token}"},
    ).json()



"""GeneWeaver Client configuration module."""
from typing import Optional, Type

from pydantic import BaseSettings, validator


class Settings(BaseSettings):
    """Settings class for GeneWeaver Client."""

    AUTH_DOMAIN = "geneweaver.auth0.com"
    AUTH_CLIENT_ID = "tvcvnbBJQr15jFds8ZkwgHoqRJ0IuGJC"
    AUTH_ALGORITHMS = ["RS256"]
    AUTH_SCOPES = ["openid", "profile", "email"]

    API_HOST: str = "https://geneweaver.org"
    API_V1_PATH: str = "/api/v1"
    API_V2_PATH: str = "/api/v2"
    API_V3_PATH: str = "/api/v3"
    API_PATH: Optional[dict] = None

    def api_v3_path(self):
        return self.API_HOST + self.API_V3_PATH

    @validator("API_PATH")
    def validate_api_path(
        cls: Type["ClientSettings"], v: Optional[dict], values: dict  # noqa: N805
    ) -> dict:
        """Construct the API URL if not explicitly set."""
        if not v:
            return {
                "v1": values["API_HOST"] + values["API_V1_PATH"],
                "v2": values["API_HOST"] + values["API_V2_PATH"],
                "v3": values["API_HOST"] + values["API_V3_PATH"],
            }
        return v

    API_URL: Optional[str] = None

    class Config:
        """Settings configuration."""

        env_file = ".env"
        env_prefix = "GW_CLIENT_"
